#!/usr/bin/env python

'''Blocked table format.

Usage:
 import blocked_table
 tb = blocked_table.TableBuilder('/tmp/mytable')
 for i in range(1000000):
   tb.add(str(i), 'Value %d' % i)
 del tb

 t = blocked_table.open('/tmp/mytable')
 for k, v in t: print k
 v = t.lookup('this is a key')
'''


from struct import pack, unpack
import bisect
import json
import logging
import os
import re
import zlib

__all__ = [ 'Table', 'TableBuilder', 'dictionary_to_table' ]

class NullCompressor(object):
  @staticmethod
  def compress(str): return str

  @staticmethod
  def decompress(str): return str


COMPRESS_LIB = { 'zlib' : zlib,
                 'none' : NullCompressor }

try:
  import snappy
  COMPRESS_LIB['snappy'] = snappy
except ImportError:
  pass

BLOCKSIZE = 128

DECODER = json.JSONDecoder()
ENCODER = json.JSONEncoder(ensure_ascii=True)

def decode(str):
  return DECODER.decode(str)

def encode(str):
  return ENCODER.encode(str)

def open_or_create(target):
  os.system('mkdir -p "%s"' % (os.path.dirname(target)))
  return open(target, 'w+')

def dictionary_to_table(dict, target):
  '''Convenience method: sort and write 'dict' to table 'target'.'''
  builder = TableBuilder(target)
  for k, v in sorted(dict.items()):
    builder.add(k, v)

def read_int(f):
  packed_val = f.read(4)
  if len(packed_val) != 4:
    return None
  return unpack('I', packed_val)[0]

def read_str(f, decompress=None):
  clen = read_int(f)
  if clen is None:
    return None
  cs = f.read(clen)
  if not decompress:
    return cs
  return decompress(cs)

def skip_str(f):
  clen = read_int(f)
  f.seek(f.tell() + clen)

def write_str(f, s, compress=None):
  if compress:
    cs = compress(s)
  else:
    cs = s
  f.write(pack('I', len(cs)))
  f.write(cs)

class BlockedTableException(Exception):
  pass


class Table(object):
  def __init__(self, f):
    self.data = open(f, 'rb')
    with open(f + '.index', 'rb') as index_file:
      header = decode(read_str(index_file))
      self.compression_format = header['compression']
      if not self.compression_format in COMPRESS_LIB:
        raise BlockedTableException, ('Compression format %s not available.' %
                                      self.compression_format)
      self.decompress = COMPRESS_LIB[self.compression_format].decompress
      index = decode(read_str(index_file))

    self.idx_keys = [str(k) for k, _ in index]
    self.idx_vals = [v for _, v in index]
    self.name = f

  def lookup(self, lk):
    block = bisect.bisect_left(self.idx_keys, lk)

    if block == len(self.idx_keys) or self.idx_keys[block] != lk:
      block -= 1

    pos = self.idx_vals[block]
    self.data.seek(pos)
    #logging.info('Lookup %s; from %s, %d', lk, self.idx_keys[block], pos)

    block = decode(read_str(self.data, self.decompress))
    for k, v in block:
      if k > lk:
        return None

      if k == lk:
        return v

    return None

  def iteritems(self):
    return self

  def __iter__(self):
    self.data.seek(0)
    while 1:
      block_str = read_str(self.data, self.decompress)
      if block_str is None:
        raise StopIteration

      block = json.loads(block_str)
      for k, v in block:
        yield (k, v)


class TableBuilder(object):
  def __init__(self, fname, compression='snappy'):
    if compression is None:
      compression = 'none'

    if not compression in COMPRESS_LIB:
      raise BlockedTableException, 'Compressor %s not available.' % compression
    self.compression = COMPRESS_LIB[compression].compress
    self.compression_format = compression

    self.data = open_or_create(fname)
    self.data.seek(0)
    self.index = []
    self.block = []

  def __del__(self):
    logging.info('Writing index for table %s', self.data.name)
    self.flush_block()
    with open(self.data.name + '.index', 'w') as index_file:
      write_str(index_file,
                encode({ 'compression' : self.compression_format }))
      write_str(index_file, encode(self.index))
    del self.data

  def flush_block(self):
    if not self.block:
      return
    key = self.block[0][0]
    #logging.debug('Adding index: (%s, %s)', key, self.data.tell())

    self.index.append((key, self.data.tell()))
    write_str(self.data,
              encode(self.block),
              self.compression)
    self.block = []

  def add(self, key, value):
    if len(self.block) > BLOCKSIZE:
      self.flush_block()

    self.block.append((key, value))
